from .racing_code import RacingCode as RacingCode
from _typeshed import Incomplete
from peak_utility.number import RepresentationalInt

class RaceGrade(RepresentationalInt):
    REGEX: str
    PHRASES: Incomplete
    def __new__(cls, grade: str | int | None, racing_code: RacingCode | None = None): ...
    def __bool__(self) -> bool: ...
    def __hash__(self): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def __lt__(self, other): ...
    def __gt__(self, other): ...



from .going_description import GoingDescription as GoingDescription

class AWGoingDescription(GoingDescription):
    SLOW: int
    STANDARD_TO_SLOW: int
    STANDARD: int
    STANDARD_TO_FAST: int
    FAST: int
    SLW = SLOW
    STS = STANDARD_TO_SLOW
    STD = STANDARD
    STF = STANDARD_TO_FAST
    FST = FAST



from peak_utility.enumeration.parsing_enum import ParsingEnum

class ObstacleStyle(ParsingEnum):
    HURDLE: int
    PLAIN_FENCE: int
    OPEN_DITCH: int
    WATER_JUMP: int
    SPECIALIST: int
    PLAIN = PLAIN_FENCE
    DITCH = OPEN_DITCH
    WATER = WATER_JUMP



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore


class Hemisphere(ParsingEnum):
    """
    An enumeration representing the geographical hemisphere.

    """

    NORTHERN = 1
    SOUTHERN = 2

    # Abbreviations
    N = NORTHERN
    S = SOUTHERN
    NORTH = NORTHERN
    SOUTH = SOUTHERN



import pendulum

from .disaster import Disaster
from .finishing_position import FinishingPosition
from .horselength import Horselength
from .outcome import Outcome


class RacePerformance:
    """
    A class for grouping together race performance stats into a single object.

    """

    def __init__(
        self,
        outcome: str | int | Disaster | FinishingPosition | Outcome,
        *,
        official_position: str | int | FinishingPosition | None = None,
        beaten_distance: str | int | Horselength | None = None,
        time: pendulum.Duration | None = None,
        comments: str | None = None,
    ):
        """
        Initialize a RacePerformance instance.

        Args:
            outcome: A disaster or finishing position
            official_position: The official finishing position
            beaten_distance: The beaten distance
            comments: Race comments on this performance

        Raises:
            ValueError: If both a disaster and a finishing position are given

        """
        self.comments = comments
        self.outcome = Outcome(outcome) if not isinstance(outcome, Outcome) else outcome
        self.official_position = (
            Outcome(official_position)
            if official_position
            else self.outcome
            if self.outcome.is_completion
            else None
        )
        self.beaten_distance = (
            None if beaten_distance is None else Horselength(beaten_distance)
        )
        self.time = time

        if not self.is_completion:
            if self.official_position:
                raise ValueError(
                    f"Cannot have both a disaster {self.outcome} and a position {self.official_position}"
                )
            if self.beaten_distance:
                raise ValueError(
                    f"Cannot have both a disaster {self.outcome} and a beaten distance {self.beaten_distance}"
                )

    def __repr__(self):
        official_position_repr = (
            f", placed {int(self.official_position._value)}"
            if self.official_position and self.official_position != self.outcome
            else ""
        )
        return f"<RacePerformance: {(int if self.is_completion else str)(self.outcome._value)}{official_position_repr}>"

    def __str__(self):
        official_position_str = (
            f", placed {self.official_position}"
            if self.official_position and self.official_position != self.outcome
            else ""
        )
        return f"{self.outcome}{official_position_str}"

    def __lt__(self, other):
        if not isinstance(other, RacePerformance):
            return NotImplemented
        return self.outcome < other.outcome

    @property
    def is_completion(self) -> bool:
        return self.outcome.is_completion

    @property
    def is_official_win(self) -> bool:
        return (
            self.is_completion
            and self.official_position is not None
            and self.official_position.is_win
        )

    @property
    def is_win(self) -> bool:
        return self.outcome.is_win



import pendulum
from .breed import Breed as Breed
from .country import Country as Country
from .horse_age import HorseAge as HorseAge
from _typeshed import Incomplete

class Horse:
    REGEX: Incomplete
    name: Incomplete
    breed: Incomplete
    country: Incomplete
    age: HorseAge | None
    def __init__(self, name: str, country: Country | str | None = None, age_or_yob: int | None = None, *, context_date: pendulum.DateTime | None = None) -> None: ...



from .age_restriction import AgeRestriction as AgeRestriction
from peak_utility.enumeration.parsing_enum import ParsingEnum

class AgeCategory(ParsingEnum):
    JUVENILE: int
    VETERAN: int
    VETERANS = VETERAN
    def to_age_restriction(self) -> AgeRestriction: ...



from peak_utility.enumeration.parsing_enum import ParsingEnum

class Handedness(ParsingEnum):
    UNKNOWN: int
    LEFT: int
    RIGHT: int
    BOTH: int
    NEITHER = UNKNOWN
    L = LEFT
    R = RIGHT
    LR = BOTH
    LH = LEFT
    RH = RIGHT



from peak_utility.enumeration.parsing_enum import ParsingEnum

class Breed(ParsingEnum):
    THOROUGHBRED: int
    ARABIAN: int
    QUARTER_HORSE: int
    AQPS: int
    TB = THOROUGHBRED
    QH = QUARTER_HORSE



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore

from .age_restriction import AgeRestriction


class AgeCategory(ParsingEnum):
    """
    An enumeration that represents the age category of a horse.

    """

    JUVENILE = 1
    VETERAN = 2

    # Alternatives
    VETERANS = VETERAN

    def to_age_restriction(self) -> AgeRestriction:
        """
        Implied age restriction, based on the category name.

        Returns:
            An AgeRestriction instance representing the age restriction of the category.

        """

        return AgeRestriction({"JUVENILE": "4yo", "VETERAN": "10yo+"}[self.name])



from peak_utility.enumeration.parsing_enum import ParsingEnum

class StallsPosition(ParsingEnum):
    INSIDE: int
    CENTRE: int
    OUTSIDE: int
    FAR = INSIDE
    MIDDLE = CENTRE
    NEAR = OUTSIDE



from .going import Going as Going
from .race_designation import RaceDesignation as RaceDesignation
from .race_distance import RaceDistance as RaceDistance
from .race_level import RaceLevel as RaceLevel
from .racecourse import Racecourse as Racecourse
from .stalls_position import StallsPosition as StallsPosition
from dataclasses import dataclass
from pendulum import DateTime as DateTime

@dataclass(kw_only=True, frozen=True)
class RaceConditions:
    datetime: DateTime
    racecourse: Racecourse
    distance: RaceDistance
    going: Going
    race_designation: RaceDesignation
    race_level: RaceLevel
    stalls_position: StallsPosition | None = ...



from horsetalk.quantity import HorsetalkQuantity


class HorseHeight(HorsetalkQuantity):
    """
    A class for measuring a horse's height.

    """

    def __str__(self) -> str:
        """
        Returns:
            A string representation of the HorseHeight object.
        """
        return f"{self.to('hh').magnitude}hh"



from peak_utility.enumeration.parsing_enum import ParsingEnum

class Surface(ParsingEnum):
    TURF: int
    DIRT: int
    ALL_WEATHER: int
    POLYTRACK: int
    TAPETA: int
    FIBRESAND: int
    T = TURF
    D = DIRT
    AW = ALL_WEATHER



from typing import Self

import pendulum
from pendulum import DateTime, Interval

from .hemisphere import Hemisphere


class HorseAge:
    """
    A class representing the age of a horse.

    Attributes:
        _official_dob (DateTime): The official date of birth of the horse.
        _actual_dob (DateTime): The actual date of birth of the horse.
        _context_date (DateTime): The context date for calculating the age of the horse.

    """

    def __init__(
        self,
        official_age: int | None = None,
        *,
        foaling_date: DateTime | None = None,
        birth_year: int | None = None,
        context_date: DateTime | None = None,
        hemisphere: Hemisphere = Hemisphere.NORTHERN,
    ):
        """
        Initializes a new instance of the HorseAge class with the specified parameters.
        If official_age and either foaling_date or birth_year are both provided, a ValueError is raised.
        If no arguments are provided, a ValueError is raised.

        Args:
            official_age (int | None): The official age of the horse.
            foaling_date (DateTime | None): The actual date of birth of the horse.
            birth_year (int | None): The birth year of the horse.
            context_date (DateTime | None): The context date for calculating the age of the horse.
            hemisphere (Hemisphere): The hemisphere where the horse was born. This is used to
                determine the horse's age based on the hemisphere-specific
                standard. The default is Hemisphere.NORTHERN.

        Returns:
            None

        Raises:
            ValueError: If invalid arguments are provided.

        """
        if official_age is not None and (foaling_date or birth_year):
            raise ValueError(
                "Cannot initialize HorseAge with both official_age and keyword"
            )

        if not (official_age or foaling_date or birth_year):
            raise ValueError(
                "Cannot initialize HorseAge without official_age, foaling_date, or birth_year"
            )

        self._actual_dob = foaling_date
        self._context_date = context_date

        year = birth_year or (
            self._base_date.year - official_age
            if official_age
            else foaling_date.year
            if foaling_date
            else None
        )

        assert year

        official_birth_month = 1 if hemisphere == Hemisphere.NORTHERN else 8

        known_born_before_august = foaling_date and foaling_date.month < 8
        current_date_before_august = self._base_date.month < 8
        year_adjustment = int(
            bool(
                hemisphere == Hemisphere.SOUTHERN
                and (
                    known_born_before_august
                    or (not foaling_date and current_date_before_august)
                )
            )
        )

        self._official_dob = pendulum.datetime(
            year - year_adjustment, official_birth_month, 1
        )

    def __repr__(self) -> str:
        """
        Returns:
            A representation of the HorseAge object.
        """
        dob_repr = (
            self._actual_dob.format("D/M/YYYY") if self._actual_dob else "unknown dob"
        )
        return f"<HorseAge: {self.official.years} ({dob_repr})>"

    def __str__(self) -> str:
        """
        Returns:
            A string representation of the HorseAge object.
        """
        return str(self.official.years)

    @property
    def official(self) -> Interval:
        """
        Calculate the official age of the horse based on its birth year or foaling date.

        Raises:
            ValueError: if the official date of birth is unknown

        Returns:
            A Interval object representing the horse's official age in years, months, and days.
        """
        return self._calc_age(self._official_dob)

    @property
    def actual(self) -> Interval:
        """
        Calculate the actual age of the horse based on its actual date of birth.

        Raises:
            ValueError: if the actual date of birth is unknown

        Returns:
            A Interval object representing the horse's actual age in years, months, and days.
        """
        if not self._actual_dob:
            raise ValueError("Cannot calculate actual age as actual dob is unknown")
        return self._calc_age(self._actual_dob)

    @property
    def date_of_birth(self) -> DateTime:
        """
        Get the date of birth of the horse.

        Returns:
            A DateTime object representing the date of birth of the horse.
        """
        if not self._actual_dob:
            raise ValueError("Actual dob is unknown")
        return self._actual_dob

    @property
    def year_of_birth(self) -> int:
        """
        Get the year of birth of the horse.

        Returns:
            An integer representing the year of birth of the horse.
        """
        return self._official_dob.year

    @property
    def _base_date(self) -> DateTime:
        """
        Get the base date for calculating the age of the horse.

        Returns:
            A DateTime object representing the current date if the context date is not set,
            otherwise the context date.
        """
        return self._context_date or pendulum.now()

    def at(self, date: DateTime) -> Self:
        """
        Set the context date for calculating the age of the horse.

        Args:
            date: A DateTime object representing the date to set the context to.

        Returns:
            A HorseAge object with the context date set.
        """
        self._context_date = date
        return self

    def _calc_age(self, dob: DateTime) -> Interval:
        """
        Calculate the age of the horse based on its date of birth.

        Args:
            dob: A DateTime object representing the date of birth of the horse.

        Returns:
            A Interval object representing the age of the horse in years, months, and days.
        """
        return (
            self._base_date - dob
            if dob < self._base_date
            else self._base_date - self._base_date
        )



from __future__ import annotations

from datetime import date
from datetime import datetime
from datetime import time
from typing import NamedTuple

class Duration:
    years: int = 0
    months: int = 0
    weeks: int = 0
    days: int = 0
    remaining_days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0
    remaining_seconds: int = 0
    microseconds: int = 0

class PreciseDiff(NamedTuple):
    years: int
    months: int
    days: int
    hours: int
    minutes: int
    seconds: int
    microseconds: int
    total_days: int

def parse_iso8601(
    text: str,
) -> datetime | date | time | Duration: ...
def days_in_year(year: int) -> int: ...
def is_leap(year: int) -> bool: ...
def is_long_year(year: int) -> bool: ...
def local_time(
    unix_time: int, utc_offset: int, microseconds: int
) -> tuple[int, int, int, int, int, int, int]: ...
def precise_diff(d1: datetime | date, d2: datetime | date) -> PreciseDiff: ...
def week_day(year: int, month: int, day: int) -> int: ...



from .disaster import Disaster as Disaster
from .finishing_position import FinishingPosition as FinishingPosition
from typing import Self

class Outcome:
    def __init__(self, value: int | str | Disaster | FinishingPosition) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __lt__(self, other: Self) -> bool: ...
    def __le__(self, other: Self) -> bool: ...
    def __gt__(self, other: Self) -> bool: ...
    def __ge__(self, other: Self) -> bool: ...
    @property
    def is_completion(self) -> bool: ...
    @property
    def is_win(self) -> bool: ...



from peak_utility.enumeration.parsing_enum import ParsingEnum

class CoatColour(ParsingEnum):
    BAY: int
    BLACK: int
    CHESTNUT: int
    DARK_BAY: int
    GREY: int
    PALOMINO: int
    WHITE: int
    ROAN: int
    B = BAY
    BL = BLACK
    C = CHESTNUT
    CH = CHESTNUT
    DB = DARK_BAY
    BR = DARK_BAY
    G = GREY
    GR = GREY
    GRAY = GREY
    P = PALOMINO
    W = WHITE
    R = ROAN
    RO = ROAN



from .going_description import GoingDescription as GoingDescription

class TurfGoingDescription(GoingDescription):
    HEAVY: int
    SOFT: int
    YIELDING_TO_SOFT: float
    GOOD_TO_SOFT: int
    YIELDING: int
    GOOD_TO_YIELDING: float
    GOOD: int
    GOOD_TO_FIRM: int
    FIRM: int
    HARD: int
    V = HEAVY
    HV = HEAVY
    HVY = HEAVY
    S = SOFT
    SFT = SOFT
    D = GOOD_TO_SOFT
    GS = GOOD_TO_SOFT
    YS = YIELDING_TO_SOFT
    Y = YIELDING
    YLD = YIELDING
    GY = GOOD_TO_YIELDING
    G = GOOD
    GD = GOOD
    GF = GOOD_TO_FIRM
    M = GOOD_TO_FIRM
    F = FIRM
    FM = FIRM
    FRM = FIRM
    HD = HARD
    HRD = HARD



from horsetalk.quantity import HorsetalkQuantity


class RaceDistance(HorsetalkQuantity):
    """
    A convenience class for representing the distance over which a race is run.

    """

    POSSIBLE_DECIMAL = r"?:(\d+(?:\.\d+)?)"
    REGEX = rf"({POSSIBLE_DECIMAL}\s?(?:m(?:iles*)?)\s*)?({POSSIBLE_DECIMAL}\s?(?:f(?:urlongs*)?)\s*)?({POSSIBLE_DECIMAL}\s?(?:y(?:ards*)?)\s*)?"

    def __str__(self) -> str:
        """
        Returns the distance as a string.
        """
        return " ".join([
            f"{int(x)}m" if (x := self.to("mile").magnitude // 1) else "",
            f"{int(x)}f" if (x := self.to("f").magnitude % 8) else "",
            f"{int(x)}y" if (x := self.to("y").magnitude % 220) else "",
        ]).strip()

    @classmethod
    def _string_arg_handler(cls, parts):
        m, f, y = parts

        if float(m or 0) > 10:
            args = (float(m or 0), "metre")
        else:
            yards = float(m or 0) * 1760 + float(f or 0) * 220 + float(y or 0)
            args = (yards, "yard")

        return args



from decimal import Decimal
from enum import Enum
from typing import Self

class Horselength(Decimal):
    class Description(Enum):
        DEAD_HEAT = 0
        NOSE = 0.05
        SHORT_HEAD = 0.1
        HEAD = 0.2
        SHORT_NECK = 0.25
        NECK = 0.3
        DISTANCE = 100
        DHT = DEAD_HEAT
        NS = NOSE
        NSE = NOSE
        SHD = SHORT_HEAD
        HD = HEAD
        SNK = SHORT_NECK
        NK = NECK
        DIST = DISTANCE
        DST = DISTANCE
    def __new__(cls, value: float | Decimal | str | None = None) -> Self: ...
    def __add__(self, other: Decimal | int) -> Horselength: ...
    def __radd__(self, other: Decimal | int) -> Horselength: ...
    def __sub__(self, other: Decimal | int) -> Horselength: ...
    def __rsub__(self, other: Decimal | int) -> Horselength: ...



from .handedness import Handedness as Handedness
from .racecourse_contour import RacecourseContour as RacecourseContour
from .racecourse_shape import RacecourseShape as RacecourseShape
from .racecourse_style import RacecourseStyle as RacecourseStyle
from .surface import Surface as Surface
from _typeshed import Incomplete

class Racecourse:
    name: Incomplete
    surface: Incomplete
    handedness: Incomplete
    contour: Incomplete
    shape: Incomplete
    style: Incomplete
    def __init__(self, name: str, surface: Surface | str, **kwargs) -> None: ...



from peak_utility.enumeration.parsing_enum import ParsingEnum

class JockeyExperienceLevel(ParsingEnum):
    AMATEUR: int
    CONDITIONAL: int
    APPRENTICE: int
    PROFESSIONAL: int
    AMATEURS = AMATEUR
    CONDITIONALS = CONDITIONAL
    APPRENTICES = APPRENTICE
    PROFESSIONALS = PROFESSIONAL



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore

from .sex import Sex


class Gender(ParsingEnum):
    """
    An enumeration representing the gender of a horse.

    """

    FOAL = 0
    YEARLING = 1
    COLT = 2
    FILLY = 3
    STALLION = 4
    MARE = 5
    GELDING = 6
    RIG = 7

    # Abbreviations
    C = COLT
    F = FILLY
    S = STALLION
    M = MARE
    G = GELDING
    R = RIG

    # Plural
    FOALS = FOAL
    YEARLINGS = YEARLING
    COLTS = COLT
    FILLIES = FILLY
    STALLIONS = STALLION
    MARES = MARE
    GELDINGS = GELDING
    RIGS = RIG

    @property
    def sex(self):
        """
        Get the sex of the horse based on its gender.

        Raises:
            ValueError: If the gender of the horse is not specific enough to determine its sex.

        Returns:
            Sex: The sex of the horse, either `Sex.FEMALE` or `Sex.MALE`.

        """
        if self in [Gender.FOAL, Gender.YEARLING]:
            raise ValueError("Not enough information to provide sex of horse")

        return Sex.FEMALE if self in [Gender.FILLY, Gender.MARE] else Sex.MALE

    @staticmethod
    def determine(official_age: int, sex: Sex | None = None, **kwargs):
        """
        Determine the gender of a horse based on its sex, official age, and optional arguments.

        Args:
            official_age: The official age of the horse in years.
            sex: The sex of the horse.
            **kwargs: Additional keyword arguments that may be used to determine the gender. Accepts is_rig and is_gelded.

        Raises:
            ValueError: If a female horse is specified as a gelding or rig.

        Returns:
            Gender: The gender of the horse based on the input arguments.

        """
        if official_age <= 1:
            return {0: Gender.FOAL, 1: Gender.YEARLING}[official_age]

        testacle_status = (
            "gelding"
            if kwargs.get("is_gelded")
            else "rig"
            if kwargs.get("is_rig")
            else None
        )

        if testacle_status:
            if sex is Sex.FEMALE:
                raise ValueError(f"Female horse cannot be {testacle_status}")
            return Gender[testacle_status]  # type: ignore

        if sex is None:
            raise ValueError("Not enough information to determine gender")

        return {
            True: {Sex.MALE: Gender.COLT, Sex.FEMALE: Gender.FILLY},
            False: {Sex.MALE: Gender.STALLION, Sex.FEMALE: Gender.MARE},
        }[official_age <= 3][sex]



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore

from .surface import Surface


class GoingDescription(ParsingEnum):
    """
    A parent enumeration for more specific going descriptions. Not intended to be instantiated directly.

    """

    @property
    def surface(self):
        """
        The surface implied by the going description.

        """
        return Surface[self.__class__.__name__.replace("GoingDescription", "")]



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore


class ObstacleStyle(ParsingEnum):
    """
    An enumeration representing the style of an inividual obstacle.

    """

    HURDLE = 1
    PLAIN_FENCE = 2
    OPEN_DITCH = 3
    WATER_JUMP = 4
    SPECIALIST = 5

    # Abbreviations
    PLAIN = PLAIN_FENCE
    DITCH = OPEN_DITCH
    WATER = WATER_JUMP



import operator
from typing import Callable, Optional

from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore


class Silks:
    description: str

    class Colour(ParsingEnum):
        """An enumeration of colours that can appear on jockey silks."""

        BEIGE = 1
        BLACK = 2
        BROWN = 3
        DARK_BLUE = 4
        DARK_GREEN = 5
        EMERALD_GREEN = 6
        GREEN = 7
        GREY = 8
        LIGHT_BLUE = 9
        LIGHT_GREEN = 10
        MAROON = 11
        MAUVE = 12
        ORANGE = 13
        PINK = 14
        PURPLE = 15
        RED = 16
        ROYAL_BLUE = 17
        WHITE = 18
        YELLOW = 19

        @staticmethod
        def phrases():
            return [
                colour.replace("_", " ").lower() for colour in Silks.Colour.__members__
            ]

    class Pattern(ParsingEnum):
        """An enumeration of patterns that can appear on jockey silks."""

        ARMLETS = 1  #
        ARMLET = ARMLETS
        BRACES = 2  #
        CHECK = 3  #
        CHECKED = CHECK
        CHEVRON = 4  #
        CHEVRONS = 5  #
        CROSS_BELTS = 6  #
        CROSS_OF_LORRAINE = 7  #
        CROSS_SASHES = CROSS_BELTS
        CUFFS = 8
        DIAMOND = 9  #
        DIAMONDS = 10  #
        DIABOLO = 11  #
        DISC = 12  #
        EPAULETS = 13  #
        HALVED = 14
        HOLLOW_BOX = 15  #
        HOOP = 16  #
        HOOPS = 17  #
        HOOPED = HOOPS
        INVERTED_TRIANGLE = 18  #
        LARGE_SPOTS = 19  #
        PANEL = 20
        PLAIN = 21  #
        QUARTERED = 22  #
        SASH = 23
        SEAMS = 24  #
        SPOTS = 25  #
        STAR = 26  #
        STARS = 27  #
        STRIPE = 28  #
        STRIPES = 29  #
        STRIPED = STRIPES
        TRIPLE_DIAMOND = 30  #

        @staticmethod
        def phrases():
            return [
                pattern.replace("_", " ").lower()
                for pattern in Silks.Pattern.__members__
            ]

        @classmethod
        def body_only(cls):
            return [
                cls.BRACES,
                cls.CHEVRON,
                cls.CROSS_BELTS,
                cls.CROSS_OF_LORRAINE,
                cls.CROSS_SASHES,
                cls.DIAMOND,
                cls.DIAMONDS,
                cls.DIABOLO,
                cls.DISC,
                cls.EPAULETS,
                cls.HOLLOW_BOX,
                cls.HOOP,
                cls.INVERTED_TRIANGLE,
                cls.LARGE_SPOTS,
                cls.PANEL,
                cls.SASH,
                cls.TRIPLE_DIAMOND,
            ]

    class Element:
        def __init__(
            self,
            primary: "Silks.Colour",
            secondary: Optional["Silks.Colour"] = None,
            pattern: Optional["Silks.Pattern"] = None,
        ):
            self.primary = primary
            self.secondary = secondary
            self.pattern = pattern

        def __repr__(self):
            return f"Element(primary={self.primary}, secondary={self.secondary}, pattern={self.pattern})"

        def __eq__(self, other):
            return (
                self.primary == other.primary
                and self.pattern == other.pattern
                and self.secondary == other.secondary
            )

    @classmethod
    def parse(cls, description: str) -> "Silks":
        """Parses a description of silks into component parts.

        Args:
            description: A description of silks.

        Returns:
            A Silks object.
        """
        self = cls()
        self.description = description

        return self

    @property
    def body(self) -> "Silks.Element":
        """Returns the body of the silks.

        Returns:
            A Silks.Element object.
        """
        element = self._convert_to_element(
            self._parts_for_element(operator.itemgetter(0))
        )
        element.secondary = element.secondary or element.primary
        element.pattern = element.pattern or Silks.Pattern.PLAIN  # type: ignore
        return element

    @property
    def cap(self) -> "Silks.Element":
        """Returns the cap of the silks.

        Returns:
            A Silks.Element object.
        """
        element = self._convert_to_element(
            self._parts_for_element(lambda parts: next(p for p in parts if "cap" in p))
            if "cap" in self.description
            else []
        )
        return self._apply_defaults(element)

    @property
    def sleeves(self) -> "Silks.Element":
        """Returns the sleeves of the silks.

        Returns:
            A Silks.Element object.
        """
        element = self._convert_to_element(
            self._parts_for_element(
                lambda parts: next(
                    p
                    for p in parts
                    if "sleeves" in p or "armlets" in p or "armlet" in p
                )
            )
            if "sleeves" in self.description
            or "armlets" in self.description
            or "armlet" in self.description
            else []
        )
        return self._apply_defaults(element)

    @property
    def _clauses(self) -> list[list[str]]:
        clauses = [
            Silks._conjoin_words(clause.split(" "))
            for clause in self.description.lower()
            .replace("(", "")
            .replace(")", "")
            .replace(" on ", " ")
            .split(", ")
        ]

        to_add = {}
        for i, clause in enumerate(clauses):
            new_clause = None
            while "and" in clause:
                index = clause.index("and")
                before_is_colour = clause[index - 1] in Silks.Colour.phrases()
                after_is_colour = clause[index + 1] in Silks.Colour.phrases()
                # colour and colour
                if before_is_colour and after_is_colour:
                    del clause[index]
                # pattern and element
                # element and element
                else:
                    new_clause = clause[: index - 1] + clause[index + 1 :]
                    while (
                        sum(word in Silks.Pattern.phrases() for word in new_clause) > 1
                    ):
                        removable = next(
                            word
                            for word in new_clause
                            if word in Silks.Pattern.phrases()
                        )
                        new_clause.remove(removable)

                    clause = clause[:index]

            clauses[i] = clause
            if new_clause:
                to_add[i + 1] = new_clause

        for key in to_add:
            clauses.insert(key, to_add[key])

        return clauses

    def _parts_for_element(self, element_selector: Callable) -> list[str]:
        main_clause = element_selector(self._clauses)
        index = self._clauses.index(main_clause)
        next_clause = (
            self._clauses[index + 1]
            if index + 1 != len(self._clauses)
            and not (
                "cap" in self._clauses[index + 1]
                or "sleeves" in self._clauses[index + 1]
                or (
                    "armlets" in self._clauses[index + 1]
                    and "sleeves" not in main_clause
                )
            )
            else []
        )

        return main_clause + next_clause

    def _apply_defaults(self, element: "Silks.Element") -> "Silks.Element":
        default_secondary = (
            self.body.secondary
            if (
                element.pattern
                or self.body.pattern
                not in [*Silks.Pattern.body_only(), Silks.Pattern.PLAIN]
            )
            else self.body.primary
        )
        default_pattern = (
            Silks.Pattern.PLAIN
            if element.primary
            or self.body.pattern in Silks.Pattern.body_only()
            or ", sleeves and cap" in self.description
            else self.body.pattern
        )
        default_primary = (
            self.body.secondary
            if ", sleeves and cap" in self.description and self.body.secondary
            else self.body.primary
        )

        element.secondary = element.secondary or element.primary or default_secondary
        element.pattern = element.pattern or default_pattern  # type: ignore

        uses_solid_pattern = element.pattern not in [
            Silks.Pattern.ARMLETS,
            Silks.Pattern.CHEVRONS,
            Silks.Pattern.SPOTS,
            Silks.Pattern.STAR,
            Silks.Pattern.STARS,
            Silks.Pattern.STRIPES,
        ]

        element.primary = (
            element.primary
            if (
                element.primary
                and (uses_solid_pattern or element.secondary != element.primary)
            )
            else default_primary
        )
        return element

    def _convert_to_element(self, words: list[str]) -> "Silks.Element":
        details = [
            (
                Silks.Colour[word]  # type: ignore
                if word in Silks.Colour.phrases()
                else Silks.Pattern[word]  # type: ignore
                if word in Silks.Pattern.phrases()
                else None
            )
            for word in words
            if word != "cap" and word != "sleeves"
        ]

        details.extend([None] * (3 - len(details)))
        details.sort(
            key=lambda x: (
                isinstance(x, Silks.Pattern),
                x is None,
                isinstance(x, Silks.Colour),
            )
        )

        return Silks.Element(*details)  # type: ignore

    @staticmethod
    def _conjoin_words(words: list[str]) -> list[str]:
        first_words = [
            "dark",
            "light",
            "royal",
            "emerald",
            "cross",
            "cross of",
            "inverted",
            "large",
            "triple",
        ]

        while any(word in words for word in first_words):
            index = words.index(next(word for word in words if word in first_words))
            if words[index] == "large":
                spots_index = words.index("spots")
                words = words[:index] + words[index + 1 : spots_index] + ["large spots"]
            else:
                words = (
                    words[:index]
                    + [" ".join([words[index], words[index + 1]])]
                    + words[index + 2 :]
                )

        return words



from peak_utility.enumeration.parsing_enum import ParsingEnum

class JumpCategory(ParsingEnum):
    HURDLE: int
    STEEPLECHASE: int
    CROSS_COUNTRY: int
    H = HURDLE
    HRD = HURDLE
    HRDLE = HURDLE
    C = STEEPLECHASE
    CH = STEEPLECHASE
    CHS = STEEPLECHASE
    CHSE = STEEPLECHASE
    CHASE = STEEPLECHASE
    CC = CROSS_COUNTRY
    XC = CROSS_COUNTRY



from .age_category import AgeCategory
from .age_restriction import AgeRestriction
from .aw_going_description import AWGoingDescription
from .breed import Breed
from .coat_colour import CoatColour
from .country import Country
from .dirt_going_description import DirtGoingDescription
from .disaster import Disaster
from .draw import Draw
from .finishing_position import FinishingPosition
from .form_break import FormBreak
from .form_figures import FormFigures
from .gender import Gender
from .going import Going
from .handedness import Handedness
from .headgear import Headgear
from .hemisphere import Hemisphere
from .horse import Horse
from .horse_age import HorseAge
from .horse_experience_level import HorseExperienceLevel
from .horse_height import HorseHeight
from .horselength import Horselength
from .jockey_experience_level import JockeyExperienceLevel
from .jump_category import JumpCategory
from .obstacle_style import ObstacleStyle
from .outcome import Outcome
from .race_class import RaceClass
from .race_conditions import RaceConditions
from .race_designation import RaceDesignation
from .race_distance import RaceDistance
from .race_grade import RaceGrade
from .race_level import RaceLevel
from .race_performance import RacePerformance
from .race_title import RaceTitle
from .race_weight import RaceWeight
from .racecourse import Racecourse
from .racecourse_contour import RacecourseContour
from .racecourse_shape import RacecourseShape
from .racecourse_style import RacecourseStyle
from .racing_code import RacingCode
from .sex import Sex
from .silks import Silks
from .stalls_position import StallsPosition
from .surface import Surface
from .turf_going_description import TurfGoingDescription

__all__ = [
    "AWGoingDescription",
    "AgeCategory",
    "AgeRestriction",
    "Breed",
    "CoatColour",
    "Country",
    "DirtGoingDescription",
    "Disaster",
    "Draw",
    "FinishingPosition",
    "FormBreak",
    "FormFigures",
    "Gender",
    "Going",
    "Handedness",
    "Headgear",
    "Hemisphere",
    "Horse",
    "HorseAge",
    "HorseExperienceLevel",
    "HorseHeight",
    "Horselength",
    "JockeyExperienceLevel",
    "JumpCategory",
    "ObstacleStyle",
    "Outcome",
    "RaceClass",
    "RaceConditions",
    "RaceDesignation",
    "RaceDistance",
    "RaceGrade",
    "RaceLevel",
    "RacePerformance",
    "RaceTitle",
    "RaceWeight",
    "Racecourse",
    "RacecourseContour",
    "RacecourseShape",
    "RacecourseStyle",
    "RacingCode",
    "Sex",
    "Silks",
    "StallsPosition",
    "Surface",
    "TurfGoingDescription",
]



from .aw_going_description import AWGoingDescription as AWGoingDescription
from .dirt_going_description import DirtGoingDescription as DirtGoingDescription
from .going_description import GoingDescription as GoingDescription
from .surface import Surface as Surface
from .turf_going_description import TurfGoingDescription as TurfGoingDescription
from _typeshed import Incomplete

class Going:
    Scales: Incomplete
    description: Incomplete
    reading: Incomplete
    def __init__(self, description: str, reading: float | None = None) -> None: ...
    def __hash__(self): ...
    def __eq__(self, other: object) -> bool: ...
    @property
    def primary(self) -> GoingDescription: ...
    @property
    def secondary(self) -> GoingDescription | None: ...
    @property
    def surface(self) -> Surface: ...
    @property
    def value(self) -> float: ...
    @staticmethod
    def multiparse(description: str) -> dict[str, 'Going']: ...



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore


class RaceDesignation(ParsingEnum):
    """
    An enumeration representing the designation or type of race.

    """

    HANDICAP = 1
    CONDITIONS = 2
    MAIDEN = 3
    AUCTION = 4
    CLAIMER = 5
    SELLER = 6
    STAKES = 7

    # Abbreviations
    HCAP = HANDICAP
    AU = AUCTION
    CL = CLAIMER
    M = MAIDEN
    S = SELLER
    STKS = STAKES

    # Alternatives
    CLAIMING = CLAIMER
    SELLING = SELLER
    NURSERY = HANDICAP



import re

import pendulum

from .breed import Breed
from .country import Country
from .horse_age import HorseAge


class Horse:
    """A class used to represent a Horse.

    Attributes:
        name (str): The name of the horse.
        breed (Breed): The breed of the horse.
        country (str): The country where the horse was bred.
        age (HorseAge | None): The age of the horse.

    Args:
        name (str): The name of the horse.
        country (str, optional): The country where the horse was bred. Defaults to None.
        age_or_yob (int, optional): The age or year of birth of the horse. Defaults to None.
        context_date (datetime, optional): The context date used to calculate the age of the horse. Defaults to current date.
    """

    REGEX = re.compile(
        r"""
        (?P<name>[A-Za-z]{1}[A-Za-z ']{1,19}[A-Za-z]{1})            # Horse's name
        \s*                                                         # Optional whitespace
        (?:\((?P<country>\w+)\))?                                   # Country of origin
        \s*                                                         # Optional whitespace
        (?P<age_or_yob>\d{1,4})?                                    # Age or year of birth
    """,
        re.VERBOSE,
    )

    def __init__(
        self,
        name: str,
        country: Country | str | None = None,
        age_or_yob: int | None = None,
        *,
        context_date: pendulum.DateTime | None = None,
    ):
        """Initializes the Horse object with name, country, age_or_yob, and context_date.

        Args:
            name (str): The name of the horse.
            country (str, optional): The country where the horse was bred. Defaults to None.
            age_or_yob (int, optional): The age or year of birth of the horse. Defaults to None.
            context_date (datetime, optional): The context date used to calculate the age of the horse. Defaults to current date.
        """
        match = re.match(Horse.REGEX, name)

        if not match:
            raise ValueError(f"Invalid horse name: {name}")

        if not context_date:
            context_date = pendulum.now()

        self.name = match.group("name")
        self.breed = None if len(self.name) <= 18 else Breed.AQPS
        self.country = (
            Country[key] if (key := match.group("country") or country) else None
        )
        self.age: HorseAge | None = None

        if country and Country[country] != self.country:
            raise ValueError(
                f"Conflicting countries in name and country arguments: {country}"
            )

        if (
            age_or_yob
            and match.group("age_or_yob")
            and int(match.group("age_or_yob")) != age_or_yob
        ):
            raise ValueError(
                f"Conflicting age_or_yob in name and age_or_yob arguments: {age_or_yob}"
            )

        age_or_yob = int(match.group("age_or_yob") or age_or_yob or -1)

        kwargs = {"context_date": context_date}
        if country:
            kwargs["hemisphere"] = self.country.hemisphere

        if age_or_yob > 999:
            self.age = HorseAge(birth_year=age_or_yob, **kwargs)
        elif age_or_yob > 0:
            self.age = HorseAge(age_or_yob, **kwargs)



from dataclasses import dataclass

from pendulum import DateTime

from .going import Going
from .race_designation import RaceDesignation
from .race_distance import RaceDistance
from .race_level import RaceLevel
from .racecourse import Racecourse
from .stalls_position import StallsPosition


@dataclass(kw_only=True, frozen=True)
class RaceConditions:
    """
    A class for grouping together race conditions into a single object.

    Properties:
        datetime: The datetime of the race
        racecourse: The racecourse on which the race is run
        distance: The race distance
        going: The going of the race
        race_designation: The designation of the race, i.e. whether it is a handicap, maiden, etc.
        race_level: The level of the race, i.e. Group 1, Group 2, etc.
        stalls_position: The position of the stalls on the track

    """

    datetime: DateTime
    racecourse: Racecourse
    distance: RaceDistance
    going: Going
    race_designation: RaceDesignation
    race_level: RaceLevel
    stalls_position: StallsPosition | None = None

    def __repr__(self):
        return (
            f"<RaceConditions: datetime={self.datetime}, "
            f"racecourse={self.racecourse!r}, "
            f"distance={self.distance!s}, "
            f"going={self.going}, "
            f"race_designation={self.race_designation.name.title()}, "
            f"race_level={self.race_level.grade or self.race_level.class_}, "
            f"stalls_position={self.stalls_position}>"
        )

    def __str__(self):
        return (
            f"{self.datetime.format('D MMM YYYY, HH:mm')}, "
            f"{self.racecourse.name}, "
            f"{self.distance!s} ({self.going}), "
            f"{self.race_designation.name.title()} "
            f"{self.race_level}"
            f"{', Stalls: ' + str(self.stalls_position.name.title()) if self.stalls_position else ''}"
        )



import re
from types import MappingProxyType

from peak_utility.number import RepresentationalInt  # type: ignore

from .racing_code import RacingCode


class RaceGrade(RepresentationalInt):
    """
    A class representing the grade of a race.

    """

    REGEX = r"G(?:roup|rade|)\s*"
    PHRASES = MappingProxyType({
        RacingCode.FLAT: "group",
        RacingCode.NATIONAL_HUNT: "grade",
    })

    def __new__(cls, grade: str | int | None, racing_code: RacingCode | None = None):
        """
        Create a RaceGrade instance.

        Args:
            grade: The grade to initialize with
            racing_code: The racing code to initialize with

        Raises:
            ValueError: If the grade is not valid

        """
        grade_text = re.sub(RaceGrade.REGEX, "", str(grade or "").title())

        if grade_text.isdigit() and 1 <= int(grade_text) < 4:
            grade_value = int(grade_text)
        elif grade_text == "Listed":
            grade_value = 4
        elif int(bool(grade_text)) == 0:
            grade_value = 5
        else:
            raise ValueError(f"{grade} is not a valid RaceGrade")

        code_from_grade = {v: k for k, v in RaceGrade.PHRASES.items()}.get(
            next(
                (x for x in ["grade", "group"] if x in str(grade).lower()),
                "default",
            ),
            None,
        )

        if code_from_grade and racing_code and code_from_grade != racing_code:
            raise ValueError(
                f"{grade} conflicts with value for racing code: {racing_code.value}"
            )

        instance = super().__new__(cls, grade_value)
        instance.racing_code = code_from_grade or racing_code or RacingCode.FLAT
        return instance

    def __repr__(self):
        return f"<RaceGrade: {str(int(self)) if int(self) < 4 else str(self) or 'Ungraded'}>"

    def __str__(self):
        if super().__eq__(5):
            return ""
        if super().__eq__(4):
            return "Listed"

        return (
            f"{RaceGrade.PHRASES.get(self.racing_code, 'group').title()} {int(self)!s}"
        )

    def __bool__(self):
        return super().__ne__(5)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, RaceGrade) or (
            isinstance(other, int)
            and other < 4
            and not isinstance(other, RepresentationalInt)
        ):
            return super().__eq__(other)

        return False

    def __ne__(self, other):
        if isinstance(other, RaceGrade) or (
            isinstance(other, int)
            and other < 4
            and not isinstance(other, RepresentationalInt)
        ):
            return super().__ne__(other)

        return True

    def __lt__(self, other):
        return super().__gt__(other)

    def __gt__(self, other):
        return super().__lt__(other)



from peak_utility.enumeration.parsing_enum import ParsingEnum

class RacecourseContour(ParsingEnum):
    UNKNOWN: int
    FLAT: int
    DOWNHILL: int
    UPHILL: int
    UNDULATING: int
    DOWN = DOWNHILL
    UP = UPHILL



from .age_category import AgeCategory as AgeCategory
from .age_restriction import AgeRestriction as AgeRestriction
from .aw_going_description import AWGoingDescription as AWGoingDescription
from .breed import Breed as Breed
from .coat_colour import CoatColour as CoatColour
from .country import Country as Country
from .dirt_going_description import DirtGoingDescription as DirtGoingDescription
from .disaster import Disaster as Disaster
from .draw import Draw as Draw
from .finishing_position import FinishingPosition as FinishingPosition
from .form_break import FormBreak as FormBreak
from .form_figures import FormFigures as FormFigures
from .gender import Gender as Gender
from .going import Going as Going
from .handedness import Handedness as Handedness
from .headgear import Headgear as Headgear
from .hemisphere import Hemisphere as Hemisphere
from .horse import Horse as Horse
from .horse_age import HorseAge as HorseAge
from .horse_experience_level import HorseExperienceLevel as HorseExperienceLevel
from .horse_height import HorseHeight as HorseHeight
from .horselength import Horselength as Horselength
from .jockey_experience_level import JockeyExperienceLevel as JockeyExperienceLevel
from .jump_category import JumpCategory as JumpCategory
from .obstacle_style import ObstacleStyle as ObstacleStyle
from .outcome import Outcome as Outcome
from .race_class import RaceClass as RaceClass
from .race_conditions import RaceConditions as RaceConditions
from .race_designation import RaceDesignation as RaceDesignation
from .race_distance import RaceDistance as RaceDistance
from .race_grade import RaceGrade as RaceGrade
from .race_level import RaceLevel as RaceLevel
from .race_performance import RacePerformance as RacePerformance
from .race_title import RaceTitle as RaceTitle
from .race_weight import RaceWeight as RaceWeight
from .racecourse import Racecourse as Racecourse
from .racecourse_contour import RacecourseContour as RacecourseContour
from .racecourse_shape import RacecourseShape as RacecourseShape
from .racecourse_style import RacecourseStyle as RacecourseStyle
from .racing_code import RacingCode as RacingCode
from .sex import Sex as Sex
from .silks import Silks as Silks
from .stalls_position import StallsPosition as StallsPosition
from .surface import Surface as Surface
from .turf_going_description import TurfGoingDescription as TurfGoingDescription

__all__ = ['AWGoingDescription', 'AgeCategory', 'AgeRestriction', 'Breed', 'CoatColour', 'Country', 'DirtGoingDescription', 'Disaster', 'Draw', 'FinishingPosition', 'FormBreak', 'FormFigures', 'Gender', 'Going', 'Handedness', 'Headgear', 'Hemisphere', 'Horse', 'HorseAge', 'HorseExperienceLevel', 'HorseHeight', 'Horselength', 'JockeyExperienceLevel', 'JumpCategory', 'ObstacleStyle', 'Outcome', 'RaceClass', 'RaceConditions', 'RaceDesignation', 'RaceDistance', 'RaceGrade', 'RaceLevel', 'RacePerformance', 'RaceTitle', 'RaceWeight', 'Racecourse', 'RacecourseContour', 'RacecourseShape', 'RacecourseStyle', 'RacingCode', 'Sex', 'Silks', 'StallsPosition', 'Surface', 'TurfGoingDescription']
